import asyncio

from jeepney import new_method_call, DBusObject, message_bus
from jeepney.integrate.asyncio import connect_and_authenticate

notifications = DBusObject('/org/freedesktop/Notifications',
                           bus_name= 'org.freedesktop.Notifications',
                           interface='org.freedesktop.Notifications')

hello_msg = new_method_call(message_bus, 'Hello')


msg = new_method_call(notifications, 'Notify', 'susssasa{sv}i', (
                       'jeepney_test',  # App name
                       0,      # Not replacing any previous notification
                       '',    # Icon
                       'Hello, world!',  # Summary
                       'This is an example notification from Jeepney', # Body
                       [],    # Actions
                       {},    # Hints
                       -1,    # expire_timeout (-1 = default)
                     ))


async def send_notification():
    (t, p) = await connect_and_authenticate(bus='SESSION')

    resp = await p.send_message(msg)
    print('Notification ID:', resp.body[0])


loop = asyncio.get_event_loop()
loop.run_until_complete(send_notification())



"""Synchronous IO wrappers around jeepney
"""
import socket

from jeepney.auth import SASLParser, make_auth_external, BEGIN
from jeepney.bus import get_bus
from jeepney.low_level import Parser, HeaderFields, MessageType
from jeepney.wrappers import hello_msg, DBusErrorResponse

class DBusConnection:
    def __init__(self, sock):
        self.sock = sock
        self.parser = Parser()
        self.outgoing_serial = 0
        hello_reply = self.send_and_get_reply(hello_msg())
        self.unique_name = hello_reply.body[0]

    def send_message(self, message):
        if message.header.serial == -1:
            self.outgoing_serial += 1
            message.header.serial = self.outgoing_serial
        data = message.serialise()
        self.sock.sendall(data)

    def recv_messages(self):
        while True:
            b = self.sock.recv(4096)
            msgs = self.parser.feed(b)
            if msgs:
                return msgs

    def send_and_get_reply(self, message):
        """Send a message, wait for the reply and return it.

        This will discard any other incoming messages until it finds the reply.
        """
        self.send_message(message)
        serial = message.header.serial
        while True:
            msgs = self.recv_messages()
            for msg in msgs:
                if serial == msg.header.fields.get(HeaderFields.reply_serial, -1):
                    if msg.header.message_type is MessageType.error:
                        raise DBusErrorResponse(msg.body)
                    return msg


def connect_and_authenticate(bus='SESSION'):
    bus_addr = get_bus(bus)
    sock = socket.socket(family=socket.AF_UNIX)
    sock.connect(bus_addr)
    sock.sendall(b'\0' + make_auth_external())
    auth_parser = SASLParser()
    while not auth_parser.authenticated:
        auth_parser.feed(sock.recv(1024))
        if auth_parser.error:
            raise Exception("Authentication failed: %r" % auth_parser.error)

    sock.sendall(BEGIN)

    conn = DBusConnection(sock)
    conn.parser.buf = auth_parser.buffer
    return conn



import asyncio

from jeepney.auth import SASLParser, make_auth_external, BEGIN
from jeepney.bus import get_bus
from jeepney.low_level import Parser, HeaderFields, MessageType
from jeepney.wrappers import hello_msg

class DBusProtocol(asyncio.Protocol):
    def __init__(self):
        self.auth_parser = SASLParser()
        self.parser = Parser()
        self.outgoing_serial = 0
        self.awaiting_reply = {}
        self.authentication = asyncio.Future()
        self.unique_name = None

    def connection_made(self, transport):
        self.transport = transport
        self.transport.write(b'\0' + make_auth_external())

    def _authenticated(self):
        self.transport.write(BEGIN)
        self.authentication.set_result(True)
        self.data_received = self.data_received_post_auth
        self.data_received(self.auth_parser.buffer)

    def data_received(self, data):
        self.auth_parser.feed(data)
        if self.auth_parser.authenticated:
            self._authenticated()
        elif self.auth_parser.error:
            self.authentication.set_exception(ValueError(self.auth_parser.error))

    def data_received_post_auth(self, data):
        for msg in self.parser.feed(data):
            reply_serial = msg.header.fields.get(HeaderFields.reply_serial, -1)
            if reply_serial in self.awaiting_reply:
                self.awaiting_reply[reply_serial].set_result(msg)

    def send_message(self, message):
        if not self.authentication.done():
            raise RuntimeError("Wait for authentication before sending messages")

        fut = None
        if message.header.serial == -1:
            self.outgoing_serial += 1
            message.header.serial = self.outgoing_serial
        if message.header.message_type == MessageType.method_call:
            fut = asyncio.Future()
            self.awaiting_reply[message.header.serial] = fut
        data = message.serialise()
        self.transport.write(data)
        return fut

async def connect_and_authenticate(bus='SESSION', loop=None):
    if loop is None:
        loop = asyncio.get_event_loop()
    (t, p) = await loop.create_unix_connection(DBusProtocol, path=get_bus(bus))
    await p.authentication
    hello_reply = await p.send_message(hello_msg())
    p.unique_name = hello_reply.body[0]
    return (t, p)
